/* ----------------------------------------------------------------------
 *  Image library for GGI. (General Bitmap Module)
 *
 *  Copyright (C) 1998 Willie Daniel <gith@quicklink.net>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ----------------------------------------------------------------------
 *
 *
 *
 */

/*
gbmhelp.c - Helper functions
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stddef.h>
#include <ctype.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "gbm.h"

/***********************************************************************/
#ifdef WILLIE

byte gbm_read_byte(FILE * inf)
{
	return (fgetc(inf) & 0xff)
}

word gbm_read_word_LSB(FILE * inf)
{
	word value;

	value = ((fgetc(inf)) & 0xff);
	value |= ((fgetc(inf)) & 0xff) << 8;

	return value;
}

word gbm_read_word_MSB(FILE * inf)
{
	word value;

	value = ((fgetc(inf)) & 0xff) << 8;
	value |= ((fgetc(inf)) & 0xff);

	return value;
}

dword gbm_read_dword_LSB(FILE * inf)
{
	dword value;

	value = ((fgetc(inf)) & 0xff);
	value |= ((fgetc(inf)) & 0xff) << 8;
	value |= ((fgetc(inf)) & 0xff) << 16;
	value |= ((fgetc(inf)) & 0xff) << 24;

	return value;
}

dword gbm_read_dword_MSB(FILE * inf)
{
	dword value;

	value = ((fgetc(inf)) & 0xff) << 24;
	value |= ((fgetc(inf)) & 0xff) << 16;
	value |= ((fgetc(inf)) & 0xff) << 8;
	value |= ((fgetc(inf)) & 0xff);

	return value;
}
#endif				/* WILLIE */
/***********************************************************************/

GBM_BOOLEAN gbm_same(const char *s1, const char *s2, int n)
{
	for (; n--; s1++, s2++)
		if (tolower(*s1) != tolower(*s2))
			return GBM_FALSE;
	return GBM_TRUE;
}

const char *gbm_find_word(const char *str, const char *substr)
{
	char buf[100 + 1], *s;
	int len = strlen(substr);

	for (s = strtok(strcpy(buf, str), " \t,");
	     s != NULL;
	     s = strtok(NULL, " \t,"))
		if (gbm_same(s, substr, len) && s[len] == '\0')
			return str + (s - buf);
	return NULL;
}

const char *gbm_find_word_prefix(const char *str, const char *substr)
{
	char buf[100 + 1], *s;
	int len = strlen(substr);

	for (s = strtok(strcpy(buf, str), " \t,");
	     s != NULL;
	     s = strtok(NULL, " \t,"))
		if (gbm_same(s, substr, len))
			return str + (s - buf);
	return NULL;
}

/* Looking at this, you might think that the gbm_file_* function pointers
   could be made to point straight at the regular read,write etc..
   If we do this then we get into problems with different calling conventions
   This simplest thing to do is simply to use the following veneers. */

static int def_open(const char *fn, int mode)
{
	return open(fn, mode);
}
static int def_create(const char *fn, int mode)
{
	return open(fn, O_CREAT | O_TRUNC | mode, S_IREAD | S_IWRITE);
}
static void def_close(int fd)
{
	close(fd);
}
static long def_lseek(int fd, long pos, int whence)
{
	return lseek(fd, pos, whence);
}
static int def_read(int fd, void *buf, int len)
{
	return read(fd, buf, len);
}
static int def_write(int fd, const void *buf, int len)
{
	return write(fd, buf, len);
}

int (*gbm_file_open) (const char *fn, int mode) = def_open;
int (*gbm_file_create) (const char *fn, int mode) = def_create;
void (*gbm_file_close) (int fd) = def_close;
long (*gbm_file_lseek) (int fd, long pos, int whence) = def_lseek;
int (*gbm_file_read) (int fd, void *buf, int len) = def_read;
int (*gbm_file_write) (int fd, const void *buf, int len) = def_write;

#define	AHEAD_BUF 0x4000

typedef struct {
	byte buf[AHEAD_BUF];
	int inx, cnt;
	int fd;
} AHEAD;

AHEAD *gbm_create_ahead(int fd)
{
	AHEAD *ahead;

	if ((ahead = malloc((size_t) sizeof(AHEAD))) == NULL)
		return NULL;

	ahead->inx = 0;
	ahead->cnt = 0;
	ahead->fd = fd;

	return ahead;
}

void gbm_destroy_ahead(AHEAD * ahead)
{
	free(ahead);
}

int gbm_read_ahead(AHEAD * ahead)
{
	if (ahead->inx >= ahead->cnt) {
		ahead->cnt = gbm_file_read(ahead->fd, (char *) ahead->buf, AHEAD_BUF);
		if (ahead->cnt <= 0)
			return -1;
		ahead->inx = 0;
	}
	return (int) (unsigned int) ahead->buf[ahead->inx++];
}
