//=-- Dots.c - Fun with flying dots... ---------------------------------------=
//
//  A bunch of neeto flying dots.  A cool LibGGI demo.
//
//=---------------------------------------------------------------------------=
//  This file is copyright (c) 1997 Chris Lattner 
//    Do with it as you please, but at least let me know of the changes you 
//    make to it...  This may be freely distributed and made a part of the GGI
//    distribution.
//=---------------------------------------------------------------------------=

#include <math.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <ggi/libggi.h>

// DOTDENSITY - Number of particles to generate per frame
int DOTDENSITY=20;

// NUMSTREAMS - The number of seperate streams
int NUMSTREAMS=8;

// NUMPARTICLES - Maximum number of particles on screen
int NUMPARTICLES=10000;

ggi_visual_t *vis;
ggi_color Palette[256];

struct PtType {
  int X, Y;
};

struct ParticleSt {
  struct PtType Pos, Vel, Acc;
  int sx, sy;
  int Life;
  uint8 Alive;     /* Is this record filled?  */
  uint8 Color;
};

struct ParticleSt *Particles;
struct PtType     *Stream, *Grav, *SVel;

int UsedParticles = 0;
int FrameNum;

void SetupPalette(void) {
  int i;
  for (i = 0; i < 64; i++) {
    Palette[i].r = i*256*4;
    Palette[i].g = 0;
    Palette[i].b = i*256*4;

    Palette[i+64].r = 0; 
    Palette[i+64].g = i*256*4; 
    Palette[i+64].b = i*256*4;

    Palette[i+128].r = 0; 
    Palette[i+128].g = 0; 
    Palette[i+128].b = i*256*4;

    Palette[i+192].r = i*256*4; 
    Palette[i+192].g = i*256*4; 
    Palette[i+192].b = 0;
  }
}

int FindParticle(void) {
  int i;
  for (i = 0; i < NUMPARTICLES; i++)
    if (!Particles[i].Alive) 
      return i;
  return -1;
}

void GenerateParticles(int Number, int SourceX, int SourceY, int Color) {
  int i;

  while (Number-- > 0) {
    i = FindParticle();
    if (i == -1) return;

    UsedParticles++;

    Particles[i].Pos.X = SourceX;
    Particles[i].Pos.Y = SourceY;
    Particles[i].Vel.X = 0;
    Particles[i].Vel.Y = 0;
    Particles[i].Acc.X = (random() % 6000) - 3000;
    Particles[i].Acc.Y = (random() % 6000) - 3000;

    Particles[i].Life = (random() % 40) + 87;
    Particles[i].Alive = 1;
    Particles[i].Color = Color;
  }
}

#ifdef __sgi
#define Clamp(X,Max) ( ((X > Max) ? Max : X) < -Max ? -Max : X)
#else
inline int Clamp(int X, int Max) {
  if (X > Max) X = Max;
  else if (X < -Max) X = -Max;
  return X;
}
#endif

void DoSource(void) {
  int i, dx, dy;
  int Dist;

  if (!(FrameNum & 63)) {
    for (i = 0; i < NUMSTREAMS; i++) {
      Grav[i].X = (random() % (280 << 16)) - (140 << 16);
      Grav[i].Y = (random() % (160 << 16)) - (80 << 16);
    }
  }

  for (i = 0; i < NUMSTREAMS; i++) {
    dx = Grav[i].X-Stream[i].X;
    dy = Grav[i].Y-Stream[i].Y;

    Dist = (int)sqrt((double) (((dx*dx) >> 16) + ((dy*dy) >> 16))) /(7+i);

    if (Dist == 0) Dist = 1;
    SVel[i].X += dx / Dist;
    SVel[i].Y += dy / Dist;

    SVel[i].X = Clamp(SVel[i].X, 3 << 16);  // Don't go too fast now!!!
    SVel[i].Y = Clamp(SVel[i].Y, 3 << 16);

    Stream[i].X += SVel[i].X;
    Stream[i].Y += SVel[i].Y;

    GenerateParticles(DOTDENSITY, Stream[i].X, Stream[i].Y, (i & 3)*64);
  }
}

int main(int argc,char *argv[]) { 
  int i, j, x, y;
  int xmax, ymax;
  ggi_mode mode;

  if (argc>=2) NUMPARTICLES=atoi(argv[1]);
  if (argc>=3) NUMSTREAMS  =atoi(argv[2]);
  if (argc>=4) DOTDENSITY  =atoi(argv[3]);

  if (! (Stream   =malloc(NUMSTREAMS  *sizeof(Stream   [0]))) ) exit(1);
  if (! (Grav     =malloc(NUMSTREAMS  *sizeof(Grav     [0]))) ) exit(1);
  if (! (SVel     =malloc(NUMSTREAMS  *sizeof(SVel     [0]))) ) exit(1);

  if (! (Particles=malloc(NUMPARTICLES*sizeof(Particles[0]))) ) exit(1);

  if (ggiInit() != 0) {
	fprintf(stderr, "%s: unable to initialize libggi, exiting.\n",
		argv[0]);
	exit(1);
  }

  if ((vis=ggiOpen(NULL)) == NULL) {
	  fprintf(stderr,
		  "%s: unable to open default visual, exiting.\n",
		  argv[0]);
	  exit(1);
  }
  ggiSetInfoFlags(vis, GGIFLAG_ASYNC);
  if(ggiSetGraphMode(vis,GGI_AUTO,GGI_AUTO,GGI_AUTO,GGI_AUTO,GGI_AUTO)) {
    ggiPanic("Cannot set default mode!\n");
  }

  ggiGetMode(vis, &mode);
  xmax = mode.virt.x;
  ymax = mode.virt.y;

  SetupPalette();
  if(mode.graphtype==GT_8BIT)
	  ggiSetPaletteVec(vis, 0, 256, Palette);

  memset(Particles, 0, sizeof(Particles[0])*NUMPARTICLES);

  while (!ggiKeyPoll(vis)) {
    DoSource();
    j = 0;
    for (i = 0; (i < NUMPARTICLES) && (j < UsedParticles); i++) 
      if (Particles[i].Alive) {
        Particles[i].Vel.X += Particles[i].Acc.X;
        Particles[i].Vel.Y += Particles[i].Acc.Y;
        Particles[i].Pos.X += Particles[i].Vel.X;
        Particles[i].Pos.Y += Particles[i].Vel.Y;
        x = (Particles[i].Pos.X >> 16) + (xmax>>1);
        y = (Particles[i].Pos.Y >> 16) + (ymax>>1);
        if ((x < 0) || (x >= xmax) || (y < 0) || (y >= ymax) ||
           (Particles[i].Life == 1) ) {
          Particles[i].Alive = 0;
          UsedParticles--;
        } else {
          Particles[i].Life--;
          Particles[i].sx = x;
          Particles[i].sy = y;
          j++;
        }
      }

    j = 0;
    ggiFillscreen(vis);

    if(mode.graphtype==GT_8BIT) {
      for (i = 0; i < 256; i++)
        ggiPutPixel(vis, i, 0, i); // Draw palette

      for (i = 0; (i < NUMPARTICLES) && (j < UsedParticles); i++) 
        if (Particles[i].Alive) {
          ggiPutPixel(vis, Particles[i].sx, Particles[i].sy, 
             (Particles[i].Life >> 1) + Particles[i].Color);
          j++;
        }
    }
    else {
      for (i = 0; i < 256; i++)
        ggiPutPixel(vis, i, 0, ggiMapColor(vis, Palette+i)); // Draw palette

      for (i = 0; (i < NUMPARTICLES) && (j < UsedParticles); i++) 
        if (Particles[i].Alive) {
          ggiPutPixel(vis, Particles[i].sx, Particles[i].sy, 
             ggiMapColor(vis, Palette + (Particles[i].Life >> 1) + Particles[i].Color));
          j++;
      }
    }

    ggiFlush(vis);

    FrameNum++;
    usleep(100);
  }

  ggiGetc(vis);
  ggiClose(vis);
  ggiExit();

  return 0;
}
