/* 
   Graphics library for GGI.  Event queue management.

   Copyright (C) 1998 Jim Ursetto - jim.ursetto@ggi-project.org
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* 
 * This file contains the implementation of ev_queues and is included
 * by the event management files of the display targets:
 * display/X/events.c, display/svgalib/events.c, etc. 
 */

/************************** GGI Helpers ***************************/
#define MAX_QUEUE 256

struct _queue {
	int cnt;
	ggi_event buff[MAX_QUEUE];
	int head;
	int tail;
};

typedef struct ev_queue {
	ggi_event_mask seen;
	struct _queue queues[sizeof(ggi_event_mask)*8];
/*	int seencnt[sizeof(ggi_event_mask)*8];*/
/*	struct _queue key;
	struct _queue ptr;
	struct _queue val;*/
} ev_queue;

/* Set up an inital, blank queue if none is there
 */
static ev_queue *_ggiEvQueueSetup(ggi_visual *vis)
{
	ev_queue *evq;

	if (VIS_EVQUEUE(vis)==NULL) {
		evq = (ev_queue *)malloc(sizeof(ev_queue));
		memset(evq,0,sizeof(ev_queue));

		VIS_EVQUEUE(vis) = evq;
	} else {
		evq=(ev_queue *)(VIS_EVQUEUE(vis));
	}

	return evq;
}

static ggi_event_mask _ggiEvQueueSeen(ggi_visual *vis,ggi_event_mask mask)
{
	ev_queue *evq;
	ggi_event_mask seen;

	evq=_ggiEvQueueSetup(vis);
	seen=(evq->seen & mask);

	return seen;
}

/* Add an event. Return 0 on either success or unrecognized event type,
 * and return -1 is the buffer is full
 */
static int _ggiEvQueueAdd(ggi_visual *vis,ggi_event *ev)
{
	ev_queue *evq=_ggiEvQueueSetup(vis);
	struct _queue *qp;
	ggi_event_mask evm = 1 << ev->any.type;

	DPRINT("Add called.\n");

	/* Prevent the queueing of globally-masked-out events. */
	if (!(evm & VIS_EVMASK(vis))) return 0;

	DPRINT("Passed mask test.\n");

	qp = &(evq->queues[ev->any.type]);

	if (qp->cnt > MAX_QUEUE)  {
		return -1;
	}

	DPRINT("Queue not full.  Adding event %08x at pos %d.\n", evm,qp->cnt);

	/* Add the event, and mark that we've seen it.. */
	memcpy(&(qp->buff[qp->head]),ev,ev->any.size);
	qp->head = (qp->head + 1) % MAX_QUEUE;
	qp->cnt++;

	evq->seen |= evm;

	return 0;
}

static int _ggiEvQueueRelease(ggi_visual *vis,ggi_event *ev,ggi_event_mask mask)
{
	ev_queue *evq=_ggiEvQueueSetup(vis);
	ggi_event_mask evm;
	
	struct _queue *qp=NULL, *qp_tmp;
	ggi_event *e_min;
	int size;
	uint32 t_min, t_tmp;
	int queue;

	evm=mask & evq->seen;

	DPRINT("Release checking for event seen.\n");

	/* Got nothing.. */
	if (evm==0) {
		return 0;
	}

	/* Max timestamp.. */
	t_min=0xFFFFFFFF;   /* (uint32) -1 may be better */

	/* Get the specified event out of the queue.  If the user
	 * asks for more than one event type, return the one that
	 * has been waiting the longest.                       */

	for (mask=1, queue=0; mask; queue++, mask<<=1) {
		qp_tmp = &(evq->queues[queue]);
		if (qp_tmp->cnt && (evm & mask) &&
				(t_tmp = qp_tmp->buff[qp_tmp->tail].any.time) < t_min) {
			DPRINT("Plausible found.\n");
			qp = qp_tmp;
			t_min = t_tmp;
		}
	}

	e_min = &(qp->buff[qp->tail]);
	
	/* Process desired event */
	qp->tail = (qp->tail + 1) % MAX_QUEUE;
	if (--qp->cnt == 0) {
		evq->seen &= ~(1 << e_min->any.type);
	}
	
	DPRINT("Retrieved event %08x.\n",1<<e_min->any.type);
	memcpy(ev,e_min,e_min->any.size);
	size=ev->any.size;

	return size;
}
