/* 
   Graphics library for GGI.

   Copyright (C) 1998 Steve Cheng <elmert@ipoline.com>

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

#include "ggi-dl.h"
#include "vgavisual.h"

char *_GGIkeystate;
ggi_event_buffer *_GGIevbuf;

/* SVGAlib's signal handlers */
static int vga_signals[] = {
	SIGHUP, SIGINT, SIGQUIT, SIGILL,
	SIGTRAP, SIGIOT, SIGBUS, SIGFPE,
	SIGSEGV, SIGPIPE, SIGALRM, SIGTERM,
	SIGXCPU, SIGXFSZ, SIGVTALRM,
	SIGPROF, SIGPWR};
static struct sigaction old_signals[sizeof(vga_signals)];

static void _GGIchecksvgamodes(ggi_visual *vis)
{
	int modes=0;
	int i;

	SVGA_PRIV(vis)->availmodes =
		malloc(vga_lastmodenumber()*sizeof(svga_ggimode));

	for (i=1; i <= vga_lastmodenumber(); i++) {
		if (vga_hasmode(i)) {
			vga_modeinfo *modeinfo;
			svga_ggimode *sgmode;
			int bpp;
			modeinfo = vga_getmodeinfo(i);
			switch(modeinfo->colors) {
			case 2:
				continue; /* 1 bpp doesn't work */
				bpp = 1;
				break;
			case 16:
				continue; /* 4 bpp doesn't work */
				bpp = 4;
				break;
			case 256:
				bpp = 8;
				break;
			case 1<<15:
				bpp = 15;
				break;
			case 1<<16:
				bpp = 16;
				break;
			case 1<<24:
				if (modeinfo->bytesperpixel==3) {
					bpp = 24;
				} else {
					bpp = 32;
				}
				break;
			default:
				continue;
			}
			sgmode = SVGA_PRIV(vis)->availmodes + modes;
			modes++;
			sgmode->w = modeinfo->width;
			sgmode->h = modeinfo->height;
			sgmode->bpp = bpp;
			sgmode->gt = ggiBPP2GT(bpp);
			if ((modeinfo->flags & IS_MODEX))
				sgmode->ismodex = 1;
			else
				sgmode->ismodex = 0;
		}
	}

	if (modes==0) {
		free(SVGA_PRIV(vis)->availmodes);
		SVGA_PRIV(vis)->availmodes=NULL;
	} else {
		if ((SVGA_PRIV(vis)->availmodes =
		     realloc (SVGA_PRIV(vis)->availmodes,
			      (modes+1)*sizeof(svga_ggimode)))
		    == NULL)
			ggiPanic("no memory available!\n");
		SVGA_PRIV(vis)->availmodes[modes].bpp=0;
	}
}

/* This is exported from svgalib, we use it to check if
   we're called from GGI's svgalib wrapper. */

extern unsigned char *__svgalib_graph_mem;

int GGIdlinit(ggi_visual *vis,const char *args)
{
	int i;
	
	if (__svgalib_graph_mem == (void*) 42) {
		ggiPanic("SVGAlib target called from the SVGAlib wrapper! Terminating.\n");
	}
	VIS_PRIVATE(vis)=malloc(sizeof(struct svga_priv));
	if (VIS_PRIVATE(vis)==NULL)
		return GGI_DL_ERROR;
	
	SVGA_PRIV(vis)->nocols=0;
	SVGA_PRIV(vis)->colormap=NULL;
	
	_GGIevbuf = malloc(sizeof(ggi_event_buffer));
	if(!_GGIevbuf) return GGI_DL_ERROR;

	_GGIevbuf->buf = malloc(4096);
	if(!_GGIevbuf->buf) return GGI_DL_ERROR;
	
	_GGIevbuf->next = _GGIevbuf->priv = NULL;
	_GGIevbuf->bigbang = 0; /* FIXME */
	_GGIevbuf->size = 4096;
	_GGIevbuf->read = _GGIevbuf->write = 0;
	_GGIevbuf->overflow = 0;

	/* Save original signal handlers because SVGAlib will set its own */
	for(i = 0; i<sizeof(vga_signals); i++)
		sigaction(vga_signals[i], NULL, old_signals+i);

	if(!ggiDebugState)
		vga_disabledriverreport();

	if(vga_init()) {
		fprintf(stderr, "display-SVGAlib: vga_init() failed\n");
		return GGI_DL_ERROR;
	}
    
	/* Allow SVGAlib to put on signal and atexit handlers */
	vga_setmode(TEXT);

	VIS_FD(vis) = 1;
	VIS_SELECT_FD(vis) = keyboard_init_return_fd();

	_GGIkeystate = keyboard_getstate();
	keyboard_clearstate();
	keyboard_seteventhandler(_GGIkeyboardhandler);

	mouse_init("", vga_getmousetype(), SAMPLE_RATE);
	mouse_seteventhandler((__mouse_handler)_GGImousehandler);
	
	_GGIchecksvgamodes(vis);

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 10) {
		/* Has mode management */
		vis->opdisplay->flush=GGIflush;
		vis->opdisplay->getmode=GGIgetmode;
		vis->opdisplay->setmode=GGIsetmode;
		vis->opdisplay->checkmode=GGIcheckmode;
	}

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 15) {
		/* Has Event management */
		vis->opdisplay->eventpoll=GGIeventpoll;
		vis->opdisplay->eventread=GGIeventread;
	}

	return GGI_DL_OPDISPLAY;
}

int GGIdlcleanup(ggi_visual *vis)
{
	int i;

	keyboard_close();
	mouse_close();
	vga_setmode(TEXT);

	/* Get rid of SVGAlib's signal handlers */
	for(i = 0; i<sizeof(vga_signals); i++)
		sigaction(vga_signals[i], old_signals+i, NULL);

#ifndef ALWAYS_SEGFAULT_ON_EXIT
	if(!ggiDebugState)
		signal(SIGSEGV,_exit);
#endif

	free(_GGIevbuf->buf);
	free(_GGIevbuf);

	fflush(stderr);
	fflush(stdout);

	if (VIS_PRIVATE(vis) != NULL)
		free(VIS_PRIVATE(vis)); 

	return EOK;
}

#include "ggidlinit.h"
