/* 
   Graphics library for GGI.

   Copyright (C) 1998 Steve Cheng <steve@ggi-project.org>

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>

#include "ggi-dl.h"
#include "debug.h"
#include "tilevisual.h"
#include "../common/parsing.inc"

#define MAX_VISUALS	256	/* This is an outrage! */

int GGIdlinit(ggi_visual *vis,const char *args)
{
	struct TileHooks *tilehook;
	char target[1024];
	int sx, sy, vx, vy, n, i=0;

	if ((! args) || (! *args)) {
		fprintf(stderr, "display-tile needs the real targets as arguments.\n");
		return -1;
	}

	VIS_PRIVATE(vis) = tilehook = malloc(sizeof(struct TileHooks));
	if (! tilehook) {
		fprintf(stderr,"display-tile: Out of memory allocating TileHooks.\n");
		return GGI_DL_ERROR;
	}

	tilehook->vislist = malloc(MAX_VISUALS*sizeof(ggi_visual_t));
	tilehook->vis_origins = malloc(MAX_VISUALS*sizeof(ggi_coord));
	tilehook->vis_sizes = malloc(MAX_VISUALS*sizeof(ggi_coord));
	if (!tilehook->vislist || !tilehook->vis_origins || !tilehook->vis_sizes) {
		fprintf(stderr,"display-tile: Out of memory allocating vislist.\n");
		free(tilehook);
		return GGI_DL_ERROR;
	}

	tilehook->buf=NULL;
	tilehook->isasync=1;  /* Yes, this SHOULD be initialized to 1. */
	tilehook->flushlock=0;
	
	/* parse each visual */
	while (*args) {

		sx = sy = vx = vy = 0;

		sscanf(args, "%d,%d,%d,%d,%n", &sx, &sy, &vx, &vy, &n);
		args+=n;

		tilehook->vis_origins[i].x = sx;
		tilehook->vis_origins[i].y = sy;
		tilehook->vis_sizes[i].x = vx;
		tilehook->vis_sizes[i].y = vy;

		args = _GGIparsetarget((char *)args, target);

		if (! args) {
			fprintf(stderr,"display-tile: parsetarget error.\n");
			free(tilehook->vislist);
			free(tilehook->vis_origins);
			free(tilehook->vis_sizes);
			free(tilehook);
			return GGI_DL_ERROR;
		}

		DPRINT("display-tile: visual #%d is %s (%d,%d)[%dx%d]\n",
			i, target, sx, sy, vx, vy);

		if (! (tilehook->vislist[i]=ggiOpen(target,NULL)) ) {
			fprintf(stderr,"display-tile: Opening of target %s failed.\n", target);
			free(tilehook->vislist);
			free(tilehook->vis_origins);
			free(tilehook->vis_sizes);
			free(tilehook);
			return GGI_DL_ERROR;
		}

		/* Don't need SYNC mode */
		ggiSetInfoFlags(tilehook->vislist[i], GGIFLAG_ASYNC);

		/* check for ':' separator */
		if (*args && (*args != ':')) {
			fprintf(stderr, "display-tile: expecting ':' between targets.\n");
			free(tilehook->vislist);
			free(tilehook->vis_origins);
			free(tilehook->vis_sizes);
			free(tilehook);
			return GGI_DL_ERROR;
		}

		if (*args == ':') args++;
		i++;
	}

	tilehook->numvis=i;

	VIS_SELECT_FD(vis) = VIS_SELECT_FD(tilehook->vislist[0]);

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 10) {
		/* Has mode management */
		vis->opdisplay->getmode=GGIgetmode;
		vis->opdisplay->setmode=GGIsetmode;
		vis->opdisplay->checkmode=GGIcheckmode;
		vis->opdisplay->flush=GGIflush;
	}
	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 15) {
		/* Has event management */
		vis->opdisplay->eventpoll=GGIeventpoll;
		vis->opdisplay->eventread=GGIeventread;
	}

	return GGI_DL_OPDISPLAY;
}

int GGIdlcleanup(ggi_visual *vis)
{
	struct TileHooks *tilehook = VIS_PRIVATE(vis);
	int i;

	ggiFlush(vis);

	/* prevents a soon-to-be unloaded function from being called */
	if (!(VIS_FLAGS(vis) & GGIFLAG_ASYNC))
		signal(SYNCSIGNAL, SIG_IGN);

	if(tilehook->buf)
		free(tilehook->buf);
	if(VIS_FB_LINEAR(vis)) {
		free(VIS_FB_LINEAR(vis));
		VIS_FB_LINEAR_SIZE(vis)=0;
	}

	for(i = 0; i<tilehook->numvis; i++)
		ggiClose(tilehook->vislist[i]);

	free(tilehook->vislist);
	free(tilehook->vis_origins);
	free(tilehook->vis_sizes);
	free(tilehook);

	return EOK;
}
