/* ------------------------------------------------------------------------ 
 *	stubs.c  - part of the TrueEmu target of GGI.
 * ------------------------------------------------------------------------
 *
 *	Copyright (c) 1998  Andrew Apted  <andrew@ggi-project.org>
 *
 *	This library is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Library General Public
 *	License as published by the Free Software Foundation; either
 *	version 2, or (at your option) any later version.
 *
 *	This library is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General
 *	Public License along with this library; see the file
 *	COPYING.LGPL.  If not, write to the Free Software Foundation,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ------------------------------------------------------------------------
 */

#include "ggi-dl.h"
#include "debug.h"
#include "structs.h"

#include "trueemu.h"

#undef putc


/* The following operations update the 'dirty region', which is used to
 * limit the area blit to the parent at ggiFlush() time.
 */

int GGIputc(ggi_visual *vis, int x, int y, char c)
{
	TrueemuHook *th = VIS_PRIVATE(vis);
	
	UPDATE_MOD(th, x, y, x+7, y+7);
	
	return th->mem_op2d->putc(vis, x, y, c);
}

int GGIputs(ggi_visual *vis, int x, int y, const char *str)
{
	TrueemuHook *th = VIS_PRIVATE(vis);

	int width = strlen(str) * 8;
	
	UPDATE_MOD(th, x, y, x+width-1, y+7);
	
	return th->mem_op2d->puts(vis, x, y, str);
}

int GGIdrawpixel(ggi_visual *vis, int x, int y)
{
	TrueemuHook *th = VIS_PRIVATE(vis);
	
	UPDATE_MOD(th, x, y, x, y);
	
	return th->mem_op2d->drawpixel(vis, x, y);
}

int GGIputpixel(ggi_visual *vis, int x, int y, ggi_pixel col)
{ 
	TrueemuHook *th = VIS_PRIVATE(vis);
	
	UPDATE_MOD(th, x, y, x, y);
	
	return th->mem_op2d->putpixel(vis, x, y, col);
}

int GGIdrawline(ggi_visual *vis, int x1, int y1, int x2, int y2)
{ 
	TrueemuHook *th = VIS_PRIVATE(vis);

	int sx=MIN(x1,x2);
	int sy=MIN(y1,y2);

	int ex=MAX(x1,x2);
	int ey=MAX(y1,y2);

	UPDATE_MOD(th, sx, sy, ex, ey);

	return th->mem_op2d->drawline(vis, x1, y1, x2, y2);
}

int GGIdrawvline(ggi_visual *vis, int x, int y, int h)
{
	TrueemuHook *th = VIS_PRIVATE(vis);
	
	UPDATE_MOD(th, x, y, x, y+h-1);
	
	return th->mem_op2d->drawvline(vis, x, y, h);
}

int GGIputvline(ggi_visual *vis, int x, int y, int h, void *buffer)
{
	TrueemuHook *th = VIS_PRIVATE(vis);
	
	UPDATE_MOD(th, x, y, x, y+h-1);
	
	return th->mem_op2d->putvline(vis, x, y, h, buffer);
}


/* The following operations DON'T update the 'dirty region', instead
 * they simply perform the blit to the parent.
 */
 
int GGIdrawhline(ggi_visual *vis, int x, int y, int w)
{
	TrueemuHook *th = VIS_PRIVATE(vis);
	int err;
	
	err = th->mem_op2d->drawhline(vis, x, y, w);

	if (! err) {
		trueemuUpdate(vis, x, y, w, 1);
	}
	return err;
}

int GGIputhline(ggi_visual *vis, int x, int y, int w, void *buffer)
{ 
	TrueemuHook *th = VIS_PRIVATE(vis);
	int err;
	
	err = th->mem_op2d->puthline(vis, x, y, w, buffer);
	
	if (! err) {
		trueemuUpdate(vis, x, y, w, 1);
	}
	return err;
}

int GGIputbox(ggi_visual *vis, int x, int y, int w, int h, void *buffer)
{ 
	TrueemuHook *th = VIS_PRIVATE(vis);
	int err;
	
	err = th->mem_op2d->putbox(vis, x, y, w, h, buffer);

	if (! err) {
		trueemuUpdate(vis, x, y, w, h);
	}
	return err;
}

int GGIdrawbox(ggi_visual *vis, int x, int y, int w, int h)
{
	TrueemuHook *th = VIS_PRIVATE(vis);
	int err;
	
	err = th->mem_op2d->drawbox(vis, x, y, w, h);

	if (! err) {
		trueemuUpdate(vis, x, y, w, h);
	}
	return err;
}

int GGIfillscreen(ggi_visual *vis)
{
	TrueemuHook *th = VIS_PRIVATE(vis);
	int err;
	
	err = th->mem_op2d->fillscreen(vis);

	if (! err) {
		trueemuUpdate(vis, 0, 0, th->width, th->height);
	}
	return err;
}

int GGIcrossblit(ggi_visual *src, int sx, int sy, int sw, int sh,
                 ggi_visual *vis, int dx, int dy, int dw, int dh)
{
	TrueemuHook *th = VIS_PRIVATE(vis);
	int err;
	
	err = th->mem_op2d->crossblit(src, sx, sy, sw, sh, 
				      vis, dx, dy, dw, dh);

	if (! err) {
		trueemuUpdate(vis, dx, dy, dw, dh);
	}
	return err;
}


/* The following routine is treated specially.  We assume that copybox
 * on the parent will be faster than performing the conversion blit
 * (especially when the parent uses acceleration).
 */

int GGIcopybox(ggi_visual *vis, int x, int y, int w, int h, int nx, int ny)
{
	TrueemuHook *th = VIS_PRIVATE(vis);
	int err;
	
	err = ggiCopyBox(th->parent, x, y, w, h, nx, ny);

	if (! err) {
		return th->mem_op2d->copybox(vis, x, y, w, h, nx, ny);
	}
	return err;
}


/* The following functions are just pass-throughs to the parent visual.
*/

int GGIsetorigin(ggi_visual *vis, int x, int y)
{
	TrueemuHook *th = VIS_PRIVATE(vis);

	return ggiSetOrigin(th->parent, x, y);
}

int GGIsetsplitline(ggi_visual *vis, int y)
{
	TrueemuHook *th = VIS_PRIVATE(vis);

	return ggiSetSplitline(th->parent, y);
}

