/* 
   Graphics library for GGI.
   XF86 display target.

   Copyright (C) 1997 Steve Cheng <elmert@ipoline.com>
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
   Copyright (C) 1997 Andreas Beck - becka@uni-duesseldorf.de
            
   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>
#include <sys/fcntl.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <signal.h>

#include "ggi-dl.h"
#include "Xvisual.h"

static int _GGIgetbpp(ggi_visual *vis)
{
	XImage *bppcheckimage;
	unsigned int bits_per_pixel=0;
	
	if ((bppcheckimage=XGetImage(XLIB_PRIV(vis)->display,
				     RootWindow(XLIB_PRIV(vis)->display,
						XLIB_PRIV(vis)->screen),
				     0, 0, 1, 1, AllPlanes, ZPixmap)) != NULL) {
		bits_per_pixel=bppcheckimage->bits_per_pixel;
		XDestroyImage(bppcheckimage);
	}
	return bits_per_pixel;
}

static int dga_signals[] = {
	SIGSEGV, SIGBUS, SIGHUP, SIGFPE};
static struct sigaction old_signals[sizeof(dga_signals)];

int GGIdlinit(ggi_visual *vis,const char *args)
{
	struct Xhooks *xhook;

	/* Just dummy variables */
	int x, y;
	unsigned z;
	Window root;

	if (geteuid()) {
		fprintf(stderr, "Need root privs to do XF86DGA.\n");
	        return GGI_DL_ERROR;
	}

	GIVE_UP_ROOT;

	for(x = 0; x<sizeof(dga_signals); x++)
		sigaction(dga_signals[x], NULL, old_signals+x);

	DPRINT("X-lib starting.\n");
	VIS_PRIVATE(vis)=xhook=malloc(sizeof(struct Xhooks));
	if (VIS_PRIVATE(vis)==NULL)
		return GGI_DL_ERROR;

	xhook->cmap=0;
	xhook->XLibLock=0;
	xhook->nocols=0;
	xhook->colormap=NULL;

	DPRINT("X-lib wants display %s.\n",args);
	xhook->display = XOpenDisplay(args);
	if (xhook->display == NULL)
		return GGI_DL_ERROR;

	DPRINT("X-lib has display %s.\n",args);

	xhook->screen = DefaultScreen(xhook->display);

	VIS_SELECT_FD(vis)=VIS_FD(vis)=ConnectionNumber(xhook->display);

	XF86DGAQueryVersion(xhook->display, &x, &y);
	DPRINT("XF86DGA version %d.%d\n", x, y);
	if(x < 1) {
		fprintf(stderr, "Your XF86DGA is too old (%d.%d).\n", x, y);
		return GGI_DL_ERROR;
	}

	XF86VidModeQueryVersion(xhook->display, &x, &y);
	DPRINT("XF86VidMode version %d.%d\n", x, y);
	
	XF86DGAQueryDirectVideo(xhook->display, xhook->screen, &x);
	if (!(x && XF86DGADirectPresent)) {
		fprintf(stderr, "No direct video capability available!\n");
		return GGI_DL_ERROR;
	}

	/* Get virtual dimensions */
	XGetGeometry(xhook->display,
		RootWindow(xhook->display, xhook->screen),
		&root, &x, &y,
		&xhook->width, &xhook->height,
		(unsigned int*)&z,
		&xhook->depth);
	
	if (xhook->depth == 24)
		xhook->depth = _GGIgetbpp(vis);

	DPRINT("Virt: %dx%d, depth: %d\n", 
		xhook->width, xhook->height, xhook->depth);

	/* Get XF86VidMode modelines */
	XF86VidModeGetAllModeLines(xhook->display, xhook->screen,
		&xhook->num_modes, &xhook->modes);

	for(x = 0; x<xhook->num_modes; x++) {
		DPRINT("Found mode: %dx%d\n",
			xhook->modes[x]->hdisplay,
			xhook->modes[x]->vdisplay);
	}

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 10) {
		/* Has mode management */
		vis->opdisplay->getmode=GGIgetmode;
		vis->opdisplay->setmode=GGIsetmode;
		vis->opdisplay->checkmode=GGIcheckmode;
	}

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 15) {
		/* Has Event management */
		vis->opdisplay->eventpoll=GGIeventpoll;
		vis->opdisplay->eventread=GGIeventread;
	}

	return GGI_DL_OPDISPLAY;
}

int GGIdlcleanup(ggi_visual *vis)
{
	struct Xhooks *xhook;
	int i;

	if (VIS_FB_LINEAR(vis)!=NULL) {
		VIS_FB_LINEAR(vis)=NULL;
		VIS_FB_LINEAR_SIZE(vis)=0;
	}

	if ((xhook=VIS_PRIVATE(vis)) != NULL)
	{
		XF86DGADirectVideo(xhook->display, xhook->screen, 0);

		XUngrabPointer(xhook->display, CurrentTime);
		XUngrabKeyboard(xhook->display, CurrentTime);

		/* Get rid of DGA's signals */
		for(i = 0; i<sizeof(dga_signals); i++)
			sigaction(dga_signals[i], old_signals+i, NULL);

#ifndef ALWAYS_SEGFAULT_ON_EXIT
		/* prevent-segfault hack */
		if(!ggiDebugState)
			signal(SIGSEGV, _exit);
#endif

		if (xhook->cmap)
			XFreeColormap(xhook->display,xhook->cmap);

		if (xhook->modes) {
			XF86VidModeSwitchToMode(xhook->display, xhook->screen,
				xhook->modes[0]);

			/* Free the modelines */
			for (i = 0; i < xhook->num_modes; i++)
				if (xhook->modes[i]->privsize > 0)
					XFree(xhook->modes[i]->private);

			XFree(xhook->modes);
		}

		if (xhook->display)
			XCloseDisplay(xhook->display);

		free(VIS_PRIVATE(vis));
	}

	return EOK;
}

#include "ggidlinit.h"
