/* -----------------------------------------------------------------------------
**	GGI console driver
** -----------------------------------------------------------------------------
**
**	Copyright (c) 1995-1997 by Steffen Seeger
**
**	This implements the basic scrolling functions and console interface
**	for the console drivers for Linux-GGI. The actual terminal emulator
**	is platform independent and based on the scroll_... functions 
**	exported by this module. (See console.h and documentation for details.)
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** -----------------------------------------------------------------------------
*/

#include <ggi/maintainers.h>

#define	MAINTAINER	Emmanuel_Marty
#define	VERSION		"$Revision: 1.19 $"

#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/sys.h>
#include <linux/interrupt.h>
#include <linux/malloc.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>

#include <linux/ascii.h>

#define INCLUDE_FROM_CONSOLE_C
#define NEED_CONSOLE_DEFS
#include <ggi/kgi.h>

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
#include <linux/vmalloc.h>
#include <linux/console.h>
#include <asm/uaccess.h>
#else
extern void register_console(void (*proc)(const char *));
#endif

#define	BGBUF_SIZE	7*PAGE_SIZE	/* scroll-back buffer size */

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)

#define CONFIG_VT_CONSOLE y
#endif

/* -----------------------------------------------------------------------------
**	basic terminal routines.
** -----------------------------------------------------------------------------
**	The scroll_... functions implement basic routines that are needed by all
**	terminal emulators. They use the framebuffer abstraction to do their
**	drawings. As we have to keep the scroll-back buffer allocated even
**	when mapped on a display (to be able to save screen contents during the
**	keyboard bottom half), we can use it as well as screen buffer on
**	hardware that does not have a textmode or no scroll-back (such as the
**	MDA or braille readers. For performance reasons, we allow the display
**	and the framebuffer to go out of sync sometimes.
** -----------------------------------------------------------------------------
*/

/*	scroll_reset is called after a struct kgi_console was allocated and 
**	attached to a display (and not mapped yet!). It initializes the
**	scroller-specific state in struct kgi_console, allocates the
**	scroll-back buffer and the (background) buffer for the alternate screen.
**	(if not already done) and adjusts dev->mode to
**	match the scroll-back buffer layout. This may fail when the visible
**	area is greater than the scroll-back buffer; -ENOMEM is returned then,
**	EOK otherwise.
*/
static int scroll_reset(struct kgi_console *cons)
{
	ASSERT(FB_SIZE_X);
	ASSERT(DPP_X);
	ASSERT(DPP_Y);

	if (! BGBUF) {

		if (! (BGBUF = kgi_malloc_huge(BGBUF_SIZE))) {

			return -ENOMEM;
		}
		memset(BGBUF, 0, BGBUF_SIZE);
	}
	BGLINES = BGBUF_SIZE / (sizeof(ushort)*FB_SIZE_X);

	SIZE_X = VISIBLE_X;
	SIZE_Y = VISIBLE_Y;

	if (BGLINES < 2*SIZE_Y) {

		kgi_free_huge(BGBUF);
		BGBUF = NULL;
		return -ENOMEM;
	}

	if (BGLINES < FB_SIZE_Y) {

		/*	We don't need to use BGBUF when mapped. However, even
		**	when mapped we don't use more than we can store in
		**	BGBUF, so we have to adjust the virtual size.
		*/
		VIRTUAL_Y = BGLINES;

	} else {

		/*	We have to use the BGBUF to do scroll-back. As a result,
		**	we reduce the virtual size to equal visible size.
		*/
		VIRTUAL_Y = VISIBLE_Y;
	}

	FB_SIZE_Y = BGLINES;
	FB_SIZE = FB_SIZE_X*FB_SIZE_Y;
	FSIZE = FB_SIZE_X*SIZE_Y;

	/*	As we are not mapped yet, we setup background mode and init
	**	the modified area to be invalid.
	*/
	FB->pbuf = (void *) BGBUF;
	CLEAR_FLAG(no_hardscroll);
	SET_FLAG(splitline);

	FROM    = FB_SIZE;
	TO      = 0;
	OFFSET  = LAST_OFFSET = 0;

	return EOK;
}

/*	scroll_done() is called when a device is going to be freed
**	but before it is detached from a display. It has to free memory
**	allocated in scroll_reset().
*/
void scroll_done(struct kgi_console *cons)
{
	if (BGBUF) {

		kgi_free_huge(BGBUF);
	}
}

/*	scroll_save is called when the console device is going to be unmapped.
**	This allows to save the framebuffer contents and emulate old console
**	behavior. However, currently this only saves the framebuffer and allows
**	full-speed mode even for dumb displays (by enabling hardware scroll and
**	splitline features).
*/
static void scroll_save(struct kgi_device *dev)
{
	struct kgi_console *cons = (void *) dev;

	if (! FLAG(no_hardscroll)) {

		TRACE(printk("%s:%i: hardscroll save\n",__FILE__,__LINE__));

		fb_copy_from(FB, BGBUF, 0, FB_SIZE);

		dev->fb->pbuf = (void *) BGBUF;
		dev->fb->ops  = dev->mem;
		SET_FLAG(splitline);
		CLEAR_FLAG(no_hardscroll);
	}
}

/*	scroll_restore() has to restore the framebuffer or cause a redraw.
*/
static void scroll_restore(struct kgi_device *dev)
{
	struct kgi_console *cons = (void *) dev;

	ASSERT(dev->type == KGI_CONSOLE_DEV);
	ASSERT(dev->dpy);

	if (VIRTUAL_Y == VISIBLE_Y) {

		int org = ORIGIN - OFFSET;
		if (org < 0) {

			org += FB_SIZE_Y;
		}
		ASSERT((0 <= org && org) < FB_SIZE_Y);

		SET_FLAG(no_hardscroll);
		SET_FLAG(splitline);

		org *= FB_SIZE_X;

		TRACE(printk("%s:%i: ops = %p, using softscroll\n",
			__FILE__,__LINE__, FB->ops));

		scroll_modified(cons, org, org + FSIZE);

	} else {

		dev->fb->pbuf = dev->dpy->fb.pbuf;
		dev->fb->ops  = dev->dpy->fb.ops;

		CLEAR_FLAG(no_hardscroll);

		if (dev->dpy->set_split) {

			SET_FLAG(splitline);
			TRACE(printk("%s:%i: fb->ops = %p, hardscroll & splitline\n",
				__FILE__,__LINE__,FB->ops));

		} else {

			CLEAR_FLAG(splitline);
			TRACE(printk("%s:%i: fb->ops = %p, hardscroll only\n",
				__FILE__,__LINE__,FB->ops));
		}

		ASSERT(BGBUF);

		fb_copy_to(FB, BGBUF, 0, FB_SIZE);
	}

	scroll_update_gadgets(cons);
	scroll_sync(cons);
}

/*	scroll_sync() has to sync the display with the scroller state. The
**	scroller keeps track of the area modified. If there were any 
**	modifications since the last scroll_sync(), the area from FROM to TO
**	contained in the current displayed area needs to be updated.
*/
void scroll_sync(struct kgi_console *cons)
{
	int orig;

	ASSERT(cons);

	if (! (DEV.state & KGI_MAPPED_ON_DISPLAY)) {

		return;
	}

	orig = ORIGIN - OFFSET;
	if (orig < 0) {

		if (   FLAG(no_hardscroll)
		    || (!FLAG(no_hardscroll) && FLAG(splitline))) {

			orig += FB_SIZE_Y;

		} else {

			orig = 0;
		}
	}
	ASSERT((0 <= orig) && (orig < FB_SIZE_Y));

	/*	when debugging, mark the first line of the framebuffer
	*/
	TRACE(fb_update(FB, 0, 0x0700 + '='));

	if (FLAG(no_hardscroll)) {

		register int frm = FROM;
		register int end = TO;
		register int fb_size = FB_SIZE;

		int back = (orig *= FB_SIZE_X) + FSIZE;
		ushort *pb = BGBUF;
		kgi_pixmap *fb = &(DEV.dpy->fb);

		ASSERT(fb);
		ASSERT(pb);

		if (OFFSET || (LAST_OFFSET != OFFSET)) {

			frm = orig;
			end = back;
		}

		if ((end < frm) || (end < orig) || (back < frm)) {

			return;
		}

		if (frm < orig) {

			frm = orig;
		}

		if (back < end) {

			end = back;
		}

		if (frm < fb_size) {

			if (fb_size >= end) {

				fb_copy_to(fb, pb + frm, frm-orig, end-frm);

			} else {

				fb_copy_to(fb, pb + frm, frm-orig, fb_size-frm);
				fb_copy_to(fb, pb, fb_size-orig, end-fb_size);
			}

		} else {

			fb_copy_to(fb, pb + frm-fb_size, frm-orig, end-frm);
		}

	} else {

		if (FLAG(scrolled)) {

			CLEAR_FLAG(scrolled);

			kgi_set_origin(&DEV, 0, 0, orig*DPP_Y);

			if (FLAG(splitline)) {

				register int
				split = DPP_Y*((orig+SIZE_Y < FB_SIZE_Y)
						? SIZE_Y : FB_SIZE_Y-orig);

				kgi_set_split(&DEV, split - 1);
			}
		}
	}

	LAST_OFFSET = OFFSET;
	FROM = UINT_MAX;
	TO   = 0;
}


/*	Scroll up n lines between t(op) and b(ottom) and erase the 'new' bottom
**	area using ERASE. If n is greater than b-t, we scroll only b-t lines.
**	If bottom is more than SIZE_Y, bottom is less or equal top, or n is zero
**	we do nothing.
*/
void scroll_up(struct kgi_console *cons, ggi_uint t, ggi_uint b, ggi_uint n)
{
	register ggi_uint src, end;		/* first and last pixel to move	*/
	register ggi_uint clean, cnt;	/* first pixel/# pixels to clean*/ 
	register ggi_uint dst;		/* pixel to move to		*/
	register ggi_uint h1, h2;		/* temporary variables		*/
	register ggi_uint fb_size=FB_SIZE;	/* local copy			*/

	if ((SIZE_Y < b) || (b <= t) || !n) {

		return;
	}

	if (n > (b - t)) {

		n = b - t;
	}

	dst  = t + ORIGIN;
	end  = b + ORIGIN;
	dst *= FB_SIZE_X;
	end *= FB_SIZE_X;

	cnt  = n*FB_SIZE_X;
	src  = dst + cnt;

	ASSERT((0 <= dst) && (dst < src) && (src <= end));

	if (MODE(alt_screen) || t || (b < SIZE_Y)) {

		/* can't use hardware	*/
		scroll_modified(cons, dst, end);

		if (end <= src) {

			clean = dst;

		} else {

			clean = end - cnt;

			if (dst >= fb_size) {

				dst -= fb_size;
				src -= fb_size;
				end -= fb_size;
			}

			ASSERT(dst < fb_size);

			if (end <= fb_size) {

				fb_move(FB, dst, src, end - src);

			} else {

				if (src < fb_size) {

					h1 = fb_size - src;
					fb_move(FB, dst, src, h1);
					src += h1;
					dst += h1;
				}

				h1 = fb_size - dst;
				h2 = end - src;

				ASSERT((h1 > 0) && (h2 > 0));

				if (h1 < h2) {

					fb_move(FB, dst, src - fb_size, h1);
					fb_move(FB, 0, src - dst, h2 - h1);

				} else {

					fb_move(FB, dst, src - fb_size, h2);
				}
			}
		}

	} else {

		/* use hardware to do scrolling */

		SET_FLAG(scrolled);

		if (OFFSET) {

			OFFSET += n;
			if (OFFSET > (FB_SIZE_Y - SIZE_Y)) {

				OFFSET = 0;
			}
		}

		ORIGIN += n;
		ORG += cnt;
		WPOS += cnt;
		POS += cnt;

		if (FLAG(splitline)) {

			/* can do wrapping		*/

			if (fb_size <= POS) {

				POS -= fb_size;
			}

			if (fb_size <= ORG) {

				ORG	-= fb_size;
				WPOS    -= fb_size;
				ORIGIN -= FB_SIZE_Y;
			}

			if (FLAG(no_hardscroll)) {

				scroll_modified(cons, ORG, ORG + FSIZE);
			}

			clean = ORG + FSIZE - cnt;

		} else {

			/* hardware scroll, but no wrapping !!! check	*/

			clean = ORG + FSIZE - cnt;

			if (FB_SIZE_Y - SIZE_Y <= ORIGIN) {

				ggi_uint half = FB_SIZE_Y >> 1;

				ORIGIN	-= half;
				half	*= FB_SIZE_X;
				clean	-= half;
				WPOS	-= half;
				POS	-= half;
				ORG	-= half;

				fb_move(FB, 0, half, fb_size - half);

				scroll_modified(cons, 0, half);
			}
		}
	}

	/* clear bottom area	*/

	scroll_modified(cons, clean, clean + cnt);

	if (clean < fb_size) {

		if (fb_size >= (clean + cnt)) {

			fb_memset(FB, clean, ERASE, cnt);

		} else {

			fb_memset(FB, clean, ERASE, fb_size - clean);
			fb_memset(FB, 0, ERASE, clean + cnt - fb_size);
		}

	} else {

		fb_memset(FB, clean - fb_size, ERASE, cnt);
	}
}


/*	Scroll down n lines between t(op) and b(ottom) and clear the new top
**	area using ERASE. If n is greater than b-t, we scroll only b-t lines.
**	If bottom is more than SIZE_Y, bottom is less or equal top or n is zero
**	we do nothing.
*/
void scroll_down(struct kgi_console *cons, ggi_uint t, ggi_uint b, ggi_uint n)
{
	register ggi_uint src, dst, end;	/* first/last pixel to move	*/
	register ggi_uint cnt;		/* size of area to clean	*/
	register ggi_uint h1, h2;		/* subexpression		*/
	register ggi_uint fb_size=FB_SIZE;	/* local copy			*/

	if ((SIZE_Y < b) || (b <= t) || !n) {

		return;
	}

	if (n > (b - t)) {

		n = b - t;
	}

	end  = t + ORIGIN;
	dst  = b + ORIGIN;
	end *= FB_SIZE_X;
	dst *= FB_SIZE_X;
	cnt  = FB_SIZE_X*n;
	src  = dst - cnt;

	ASSERT((end <= src) && (src < dst));

	scroll_modified(cons, end, dst);

	if (src > end) {

		if (end >= fb_size) {

			dst -= fb_size;
			src -= fb_size;
			end -= fb_size;
		}

		if (dst <= fb_size) {

			fb_move(FB, end + cnt, end, src - end);

		} else {

			if (src > fb_size) {

				h1 = src - fb_size;
				fb_move(FB, dst - src, 0, h1);
				src -= h1;
				dst -= h1;
			}

			h1 = dst - fb_size;
			h2 = src - end;

			if (h1 < h2) {

				fb_move(FB, 0, src - h1, h1);
				h2 -= h1;
				fb_move(FB, fb_size - h2, end, h2);

			} else {

				fb_move(FB, dst - h2 - fb_size, end, h2);
			}
		}
	}

	/* clear top area */

	if (end < fb_size) {

		if (fb_size >= (end + cnt)) {

			fb_memset(FB, end, ERASE, cnt);

		} else {

			fb_memset(FB, end, ERASE, fb_size - end);
			fb_memset(FB, 0, ERASE, end + cnt - fb_size);
		}

	} else {

		fb_memset(FB, end - fb_size, ERASE, cnt);
	}
}


/*	Erase in display using ERASE. The area cleared depends on arg.
*/
void scroll_erase_display(struct kgi_console *cons, int arg)
{
	ggi_uint src, cnt = FSIZE;		/* area to erase	*/
	ggi_uint fb_size = FB_SIZE;		/* local copy		*/

	switch(arg) {

	case 0:	/* from cursor to end of screen	*/
		src = WPOS;
		cnt -= WPOS - ORG;
		break;

	case 1:	/* from start of screen to cursor */
		src = ORG;
		cnt = WPOS + 1 - ORG;
		break;

	case 2:	/* erase entire screen */
		src = ORG;
		break;

	default:
		return;
	}

	scroll_modified(cons, src, src + cnt);

	if (src < fb_size) {

		if (fb_size >= (src + cnt)) {

			fb_memset(FB, src, ERASE, cnt);

		} else {

			fb_memset(FB, src, ERASE, fb_size - src);
			fb_memset(FB, 0, ERASE, src + cnt - fb_size);
		}

	} else {

		fb_memset(FB, src - fb_size, ERASE, cnt);
	}

	CLEAR_FLAG(need_wrap);
}

/*	erase in line using ERASE. The area erased depends on arg.
*/
void scroll_erase_line(struct kgi_console *cons, int arg)
{
	ggi_uint src, cnt;

	switch (arg) {

	case 0:	/* from cursor to end of line	*/
		src = WPOS;
		cnt = FB_SIZE_X - X;
		break;

	case 1:	/* from line start to cursor	*/
		src = WPOS - X;
		cnt = X + 1;
		break;

	case 2:	/* erase entire line	*/
		src = WPOS - X;
		cnt = FB_SIZE_X;
		break;

	default:
		return;
	}

	scroll_modified(cons, src, src + cnt);

	if (FB_SIZE <= src) {

		src -= FB_SIZE;
	}

	fb_memset(FB, src, ERASE, cnt);
	CLEAR_FLAG(need_wrap);
}

/*	Insert n ERASE characters.
*/
void scroll_insert_chars(struct kgi_console *cons, ggi_uint n)
{
	ggi_uint src = WPOS;
	ggi_uint cnt = FB_SIZE_X - X;

	ASSERT((cnt > 0) && (n > 0));

	scroll_modified(cons, src, src + cnt);

	if (FB_SIZE <= src) {

		src -= FB_SIZE;
	}

	if (n < cnt) {

		fb_move(FB, src + n, src, cnt - n);
		cnt = n;
	}

	fb_memset(FB, src, ERASE, cnt);
	CLEAR_FLAG(need_wrap);
}

/*	Delete n characters. Clear the end of the line using ERASE.
*/
void scroll_delete_chars(struct kgi_console *cons, ggi_uint n)
{
	int cnt = FB_SIZE_X - X;
	int src = WPOS;

	scroll_modified(cons, src, src + cnt);

	if (FB_SIZE <= src) {

		src -= FB_SIZE;
	}

	if (n < cnt) {

		fb_move(FB, src, src + n, cnt - n);
		fb_memset(FB, src + cnt - n, ERASE, n);

	} else {

		fb_memset(FB, src, ERASE, cnt);
	}

	CLEAR_FLAG(need_wrap);
}


/*	erase the following n chars. (maximum is end of line).
*/
void scroll_erase_chars(struct kgi_console *cons, ggi_uint n)
{
	if (n > (SIZE_X - X)) {

		n = FB_SIZE_X - X;
	}

	fb_memset(FB, POS, ERASE, n);
	scroll_modified(cons, WPOS, WPOS + n);
	CLEAR_FLAG(need_wrap);
}


/*	Move cursor to the specified position. We have to check all boundaries.
*/
void scroll_gotoxy(struct kgi_console *cons, int new_x, int new_y)
{
	int max_y, min_y;

	if (MODE(origin)) {

		min_y = TOP;
		max_y = BOTTOM;

	} else {

		min_y = 0;
		max_y = SIZE_Y;
	}

	X = (new_x <     0) ?     0 : ((new_x < SIZE_X) ? new_x : SIZE_X - 1);
	Y = (new_y < min_y) ? min_y : ((new_y < max_y ) ? new_y : max_y  - 1);

	POS = WPOS = ORG + Y*FB_SIZE_X + X;

	if (FB_SIZE <= POS) {

		POS -= FB_SIZE;
	}

	CLEAR_FLAG(need_wrap);
}

/*	increase the scroll-back offset by <lines> lines. <lines> defaults to
**	half the number of lines of the visible screen if less or equal zero. 
**	As we may be called from a IRQ service, we do not sync the display
**	because this may take some long time. We just mark the CONSOLE_BH and
**	do syncing there.
*/
void scroll_backward(struct kgi_console *cons, int lines)
{
	int max = FB_SIZE_Y - SIZE_Y;

	if (lines <= 0) {

		lines = SIZE_Y/2;
	}

	OFFSET += lines;
	if (OFFSET > max) {

		OFFSET = max;
	}

	SET_FLAG(scrolled);

	scroll_need_sync(cons);
}

/*	decrease the scroll-back offset by <lines> lines. <lines> defaults to
**	half the number of lines of the visible screen if less or equal zero.
**	As we may be called from an IRQ service, we do not sync the display
**	because this may take some long time. We just mark the CONSOLE_BH
**	and do syncing there.
*/
void scroll_forward(struct kgi_console *cons, int lines)
{
	if (lines <= 0) {

		lines = SIZE_Y/2;
	}

	OFFSET -= lines;
	if (OFFSET < 0) {

		OFFSET = 0;
	}

	SET_FLAG(scrolled);

	scroll_need_sync(cons);
}


/*	Do a linefeed. This means to scroll up one line if the cursor is at
**	the bottom line of the scrolling region or simply to move the cursor
**	down one line.
*/
void scroll_lf(struct kgi_console *cons)
{
	if ((Y+1) == BOTTOM) {

		scroll_up(cons, TOP, BOTTOM, 1);

	} else {

		if (Y < (SIZE_Y-1)) {

			Y++;
			WPOS += FB_SIZE_X;
			POS  += FB_SIZE_X;

			if (FB_SIZE <= POS) {

				POS -= FB_SIZE;
			}
		}
	}

	CLEAR_FLAG(need_wrap);
}

/*	Do a reverse linefeed. This means to scroll down one line if the cursor
**	is at the top of the scrolling region or simply to move the cursor up
**	one line.
*/
void scroll_reverse_lf(struct kgi_console *cons)
{
	if (Y == TOP) {

		scroll_down(cons, TOP, BOTTOM, 1);

	} else {

		if (Y > TOP) {

			Y--;
			WPOS -= FB_SIZE_X;

			if (POS < FB_SIZE_X) {

				POS += FB_SIZE - FB_SIZE_X;

			} else {

				POS -= FB_SIZE_X;
			}
		}
	}

	CLEAR_FLAG(need_wrap);
}

struct wait_queue *keypress_wait = NULL;

/*
**	This is a printk() implementation based on the scroll_* functions. 
*/
#ifdef CONFIG_VT_CONSOLE

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
static void con_printk(const char *b, ggi_uint count)
#else
void console_print(const char *b)
#endif
{
	static int printing = 0;

	struct kgi_console *cons = (void *) kgidev[printk_console];
	ggi_uint c;
	ggi_uint old_pos;

	if (printing || !cons) {

		return;
	}

	printing = 1;

	old_pos = WPOS;

	if (DEV.state & KGI_DEVICE_IS_FOCUSED) {

		ASSERT(DEV.state & KGI_IS_MAPPED_ON_DISPLAY);
		kgi_undo_gadgets(&DEV);
	}

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
	while ((c = *(b++))) {
#else
	while (count--) {
		c = *(b++);
#endif
		if ((c == ASCII_LF) || (c == ASCII_CR) || FLAG(need_wrap)) {

			if (c != ASCII_CR) {

				scroll_modified(cons, old_pos, WPOS);
				scroll_lf(cons);
				scroll_sync(cons);
			}

			scroll_cr(cons);
			old_pos = WPOS;


			if ((c == ASCII_LF) || (c == ASCII_CR)) {
				continue;
			}
		}

		scroll_write(cons, c);
	}

	if (WPOS != old_pos) {

		scroll_modified(cons, old_pos, WPOS);
		scroll_sync(cons);
	}
	scroll_update_gadgets(cons);

	if (DEV.state & KGI_DEVICE_IS_FOCUSED) {

		ASSERT(DEV.state & KGI_MAPPED_ON_DISPLAY);
		kgi_show_gadgets(&DEV);
	}

	printing = 0;
}

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
static void con_wait_key(void)
{
	sleep_on(&keypress_wait);
}

static int con_dev(void)
{
	return MKDEV(TTY_MAJOR, printk_console + 1);
}

static struct console con_printk_console={
	con_printk,
	NULL,
	con_wait_key,
	con_dev,
	NULL
};
#endif

#endif /* #ifdef CONFIG_VT_CONSOLE */

/* -----------------------------------------------------------------------------
 *	A very dumb console parser to have output even with no parser loaded.
 * -----------------------------------------------------------------------------
 */

static void dumb_do_reset(struct kgi_console *cons, ggi_uint do_reset)
{
	DEV.event_mask = emKeyPress | emKeyRepeat;

	if (DEV.state & KGI_DEVICE_IS_FOCUSED) {

		ASSERT(DEV.state & KGI_MAPPED_ON_DISPLAY);
		kgi_undo_gadgets(&DEV);
	}

	cons->scroll.mode = 0;
	SET_MODE(show_cursor);
	SET_MODE(auto_wrap);

	if (TTY) {

		TTY->winsize.ws_col = SIZE_X;
		TTY->winsize.ws_row = SIZE_Y;
	}

	ORIGIN = ORG = 0;
	TOP    = 0;
	BOTTOM = SIZE_Y;

	BELL_PITCH = 440 /* Hz */;
	BELL_DURATION = 200 /* msec */;

	ATTRFL = ATTR_NORMAL | ATTR_COLOR(COL_LIGHTGRAY, COL_BLACK);

	scroll_update_attr(cons);
	scroll_gotoxy(cons, 0, 0);
	scroll_erase_display(cons, 2);
	scroll_update_gadgets(cons);
	scroll_sync(cons);

/* !!!	kbd_reset(&(KBD_STATE)); */

	if (DEV.state & KGI_DEVICE_IS_FOCUSED) {

		ASSERT(DEV.state & KGI_MAPPED_ON_DISPLAY);
		kgi_show_gadgets(&(DEV));
	}

	/* !!! kbd_setleds */
}

static void dumb_handle_event(struct kgi_device *dev, ggi_event *ev)
{
	#define	cons	((struct kgi_console *) dev)
	
	int sym  = ev->key.sym;
	int type = KTYP(sym);
	int val  = KVAL(sym);

	if (  !((ev->any.type == evKeyPress) || (ev->any.type == evKeyRepeat))
	    || (sym == K_VOID) || !TTY) {

		return;
	}

	if (type >= 0xF0) {

		type -= 0xF0;

		if ((type == KT_LETTER) || (type == KT_LATIN)) {

			tty_insert_flip_char(TTY, val, 0);
		}

		if (type == KT_SPEC) {

			switch (K(KT_SPEC,val)) {

				case K_HOLD:

					if (TTY->stopped) {

						start_tty(TTY);

					} else {

						stop_tty(TTY);
					}
					return;

				case K_ENTER:

					tty_insert_flip_char(TTY, ASCII_CR, 0);
					break;

				case K_SCROLLFORW:

					scroll_forward(cons, 0);
					break;

				case K_SCROLLBACK:

					scroll_backward(cons, 0);
					break;

				default:
					return;
			}
		}
	
		wake_up(&keypress_wait);
		tty_schedule_flip(TTY);
	}
	#undef	cons
}

static int dumb_do_write(struct kgi_console *cons, int from_user,
			  const char *buf, int count)
{
	ggi_uint cnt = 0;
	uchar c;
	ggi_uint old_pos = WPOS;

	while (!(TTY->stopped) && count) {

		#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)

			c = from_user ? get_user(buf) : *buf;
		#else

			if (from_user) {

				if (get_user(c, buf)) {

					return -EFAULT;
				}

			} else {

				c = *buf;
			}
		#endif

		buf++; cnt++; count--;

		if ((c == ASCII_LF) || (c == ASCII_CR) || FLAG(need_wrap)) {

			if (c != ASCII_CR) {

				scroll_modified(cons, old_pos, WPOS);
				scroll_lf(cons);
			}

			scroll_cr(cons);
			old_pos = WPOS;

			if ((c == ASCII_LF) || (c == ASCII_CR)) {

				continue;
			}
		}

		if (c == ASCII_BS) {

			scroll_modified(cons, old_pos, WPOS);
			scroll_bs(cons);
			old_pos = WPOS;
			continue;
		}

		scroll_write(cons, c);
	}

	if (old_pos != WPOS) {

		scroll_modified(cons, old_pos, WPOS);
	}

	return cnt;
}

/* -----------------------------------------------------------------------------
 *	console TTY driver
 * -----------------------------------------------------------------------------
 */

#define	DO_RESET	1

/* !!!	This should handle assignment of registered terminal emulators later.
** !!!	For now we only enter the neccessary fields for the dumb or xterm
** !!!	parser.
*/
static int con_assign_parser(struct kgi_console *cons, int do_reset)
{
	ASSERT(cons);

	DEV.restore		= &scroll_restore;
	DEV.save		= &scroll_save;

/*	more annoying than anything else - commented out.. [emarty]

	if (DEV.id == 5) {

		DEV.handle_event	= &dumb_handle_event;
		DO_WRITE		= &dumb_do_write;
		dumb_do_reset(cons, 1);

	}
	else {
*/

		DEV.handle_event	= &xterm_handle_event;
		DO_WRITE		= &xterm_do_write;
		xterm_do_reset(cons, 1);
/*
	}
*/

	return EOK;
}

static void con_bottomhalf(void)
{
	int i;

	for (i = 0; i < MAX_NR_CONSOLES; i++) {

		if (test_bit(i, con_need_sync)) {

			kgi_undo_gadgets(kgidev[i]);
			scroll_sync((struct kgi_console *) kgidev[i]);
			scroll_update_gadgets((struct kgi_console *) kgidev[i]);
			clear_bit(i, con_need_sync);
			kgi_show_gadgets(kgidev[i]);
		}
	}
}

static int con_open(struct tty_struct *tty, struct file *filp)
{
	ggi_uint kgiindex = MINOR(tty->device) - tty->driver.minor_start;
	struct kgi_console *cons = (struct kgi_console *) kgidev[kgiindex];

	if (!cons) {

		TRACE(printk(KERN_ERR "%s:%i: allocating console %i...",\
			__FILE__,__LINE__,kgiindex));

		if (!(cons = kgi_alloc_device(kgiindex, GFP_USER))) {

			TRACE(printk("failed because out of memory\n"));
			return -ENOMEM;
		}

		if (kgi_attach_device(&DEV)) {

			TRACE(printk("failed because couldn't attach\n"));
			kgi_free_device(&DEV);
			return -ENXIO;
		}

		if (scroll_reset(cons) || con_assign_parser(cons, DO_RESET)) {

			TRACE(printk("failed, because couldn't reset\n"));

			scroll_done(cons);
			kgi_detach_device(&DEV);
			kgi_free_device(&DEV);
			return -ENXIO;
		}

		if (! DEV.dpy->dev) {

			kgi_map_device(&DEV);
		}
		TRACE(printk("successful\n"));
	}

	TTY = tty;
	DEV.refcnt++;

	tty->driver_data = cons;
	if (!tty->winsize.ws_row && !tty->winsize.ws_col) {

		tty->winsize.ws_row = SIZE_Y;
		tty->winsize.ws_col = SIZE_X;
	}

	return 0;
}

static void con_close(struct tty_struct *tty, struct file *filp)
{
	struct kgi_console *cons = tty->driver_data;

	if (cons) {

		if (--(DEV.refcnt) == 1) {

			TRACE(printk("%s:%i: freeing console %i\n",\
				__FILE__,__LINE__,DEV.id));

			if (DEV.state & KGI_MAPPED_ON_DISPLAY) {

				ASSERT(DEV.dpy->dev == &DEV);
				kgi_unmap_device(&DEV);
			}
			scroll_done(cons);
			kgi_detach_device(&DEV);
			kgi_free_device(&DEV);
		}
	}
}

static void con_start(struct tty_struct *tty)
{
	struct kgi_console *cons = tty->driver_data;

/* !!!	kbd_clr_led(&KBD_STATE, KBD_LED_SCROLLOCK); **/
}

static void con_stop(struct tty_struct *tty)
{
	struct kgi_console *cons = tty->driver_data;

/* !!!	kbd_set_led(&KBD_STATE, KBD_LED_SCROLLOCK); */
}

static int con_write(struct tty_struct *tty, int from_user, 
		     const uchar *buf, int count)
{
	struct kgi_console *cons = tty->driver_data;
	int cnt;

	if (DEV.state & KGI_DEVICE_IS_FOCUSED) {

		ASSERT(DEV.state & KGI_MAPPED_ON_DISPLAY);
		kgi_undo_gadgets(&DEV);
	}

	cnt = DO_WRITE(cons, from_user, buf, count);

	scroll_sync(cons);
	scroll_update_gadgets(cons);

	if (DEV.state & KGI_DEVICE_IS_FOCUSED) {

		ASSERT(DEV.state & KGI_MAPPED_ON_DISPLAY);
		kgi_show_gadgets(&DEV);
	}

	return cnt;
}

static void con_put_char(struct tty_struct *tty, unsigned char c)
{
	struct kgi_console *cons = tty->driver_data;

	if (DEV.state & KGI_DEVICE_IS_FOCUSED) {

		ASSERT(DEV.state & KGI_MAPPED_ON_DISPLAY);
		kgi_undo_gadgets(&DEV);
	}

	DO_WRITE(cons, 0, &c, 1);
	scroll_sync(cons);
}

static void con_flush_char(struct tty_struct *tty) {
	struct kgi_console *cons = tty->driver_data;

	scroll_update_gadgets(cons);

	if (DEV.state & KGI_DEVICE_IS_FOCUSED) {

		ASSERT(DEV.state KGI_MAPPED_ON_DISPLAY);
		kgi_show_gadgets(&DEV);
	}
}

static int con_write_room(struct tty_struct *tty)
{
	return tty->stopped ? 0 : 4096;
}

static int con_chars_in_buffer(struct tty_struct *tty)
{
	return 0;
}

static int			con_refcount;
static struct tty_struct	*con_table[MAX_NR_CONSOLES];
static struct termios		*con_termios[MAX_NR_CONSOLES];
static struct termios		*con_termios_locked[MAX_NR_CONSOLES];

static struct tty_driver console_driver={0};

static void console_driver_init(void)
{
	memset(&console_driver,0,sizeof(struct tty_driver));

	console_driver.magic=TTY_DRIVER_MAGIC;
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,28)
	console_driver.driver_name="/dev/tty";
#endif
	console_driver.name="tty";
	console_driver.name_base=1;
	console_driver.major=TTY_MAJOR;
	console_driver.minor_start=1;
	console_driver.num=MAX_NR_CONSOLES;
	console_driver.type=TTY_DRIVER_TYPE_CONSOLE;
	console_driver.subtype=0;
	console_driver.flags=TTY_DRIVER_REAL_RAW;
	console_driver.refcount=&con_refcount;
	
	console_driver.table=con_table;
	console_driver.termios=con_termios;
	console_driver.termios_locked=con_termios_locked;
	console_driver.init_termios = tty_std_termios;

	console_driver.open=con_open;
	console_driver.close=con_close;
	console_driver.write=con_write;
	console_driver.put_char=con_put_char;
	console_driver.flush_chars=con_flush_char;

	console_driver.write_room=con_write_room;
	console_driver.chars_in_buffer=con_chars_in_buffer;

	console_driver.stop=con_stop;
	console_driver.start=con_start;
}

int	printk_console;
ulong	con_need_sync[(MAX_NR_CONSOLES+sizeof(ulong)*8-1)/(sizeof(ulong)*8)];

void dev_console_init(void)
{
	struct kgi_console *cons;

	if (kgidpy[0]) {	/* any display detected during boot ?	*/

		#define	PANIC(x)	panic(__FILE__": " x);

		console_driver_init();

		memset(con_need_sync, 0, sizeof(con_need_sync));
		init_bh(CONSOLE_BH, con_bottomhalf);

		if (tty_register_driver(&console_driver)) {

			PANIC("Could not register console driver!\n");
		}

		if (!(cons = kgi_alloc_device(0, GFP_KERNEL))) {

			PANIC("Could not allocate console struct!\n");
		}

		if (kgi_attach_device(&DEV)) {

			PANIC("Could not attach console device to display!\n");
		}

		if (scroll_reset(cons)) {

			PANIC("Could not reset scroller state!\n");
		}

		if (con_assign_parser(cons, DO_RESET)) {

			PANIC("Could not assign console parser!\n");
		}

		DEV.refcnt++;

		printk_console = 0;

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
		register_console(console_print);
#else
	#ifdef CONFIG_VT_CONSOLE
		register_console(&con_printk_console);
	#endif
#endif
		kgi_map_device(&DEV);
	}
}
