/* ---------------------------------------------------------------------------
** 	testvid.c - A video mode testing utility for GGI
** ---------------------------------------------------------------------------
** 
** 	Copyright (c) 1997, Jon M. Taylor [taylorj@ecs.csus.edu]
** 
** ---------------------------------------------------------------------------
** 
** 	$Log: testvid.c,v $
** 	Revision 1.8  1997/12/02 04:26:18  taylorj
** 	Minor: Reformatting and a few minor bugfixes
** 	
** 	Revision 1.7  1997/11/08 16:47:39  taylorj
** 	Minor: Fixed a few typos.
** 	
** 	Revision 1.6  1997/11/08 15:13:02  becka
** 	Namespace cleanup ... probably broke all and everything ...
** 	
** 	Revision 1.5  1997/10/21 02:51:41  taylorj
** 	Minor: Reorganized code structure and various small cleanups.
** 	
** 	Revision 1.4  1997/09/27 13:36:33  martin
** 	Small typo: Didn't notice that "c" had been changed to "color".
**
** 	Revision 1.3  1997/09/27 13:31:28  martin
** 	Fixed 15 and 16 bit test picture, and added 24 bit drawpixel.
**
** 	Revision 1.2  1997/09/24 08:07:03  taylorj
** 	Major: Huge source cleanups, rewritten event handling, program now 
**	exits on a keypress and restores the original video mode, and added 
**	CVS logging to header comments.
**
** ---------------------------------------------------------------------------
*/

#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <ggi/ggi_commands.h>
#include <ggi/IBM/modes.h>
#include <ggi/palette/332.inc>

/* Set a graphics mode */
int setmode(int fd, int xvis, int yvis, int xvirt, int yvirt, int depth)
{ 
	int err, type, i;
	ggi_mode mode;

	/* Convert bit depth to GGI mode type */
	switch (depth) 
	{
	case 1: 
		type = GT_1BIT; 
		break;
	case 4: 
		type = GT_4BIT; 
		break;
	case 8: 
		type = GT_8BIT;
		break;
	case 15: 
		type = GT_15BIT; 
		break;
	case 16: 
		type = GT_16BIT; 
		break;
	case 24: 
		type = GT_24BIT; 
		break; 
	case 32: 
		type = GT_32BIT; 
		break;
	}

	mode.frames	= 0;		/* ??? */
	mode.visible.x	= xvis;		/* Physical display width	*/
	mode.virt.x	= xvirt;	/* Virtual display width	*/
	mode.visible.y	= yvis;		/* Physical display height	*/
	mode.virt.y	= yvirt;	/* Virtual display height	*/
	mode.graphtype	= type;		/* Video mode type		*/
	mode.text.x	= 0;		/* Character cell width		*/
	mode.text.y	= 0;		/* Character cell height	*/

	if (err = ioctl(fd, DRIVER_SETGRAPHMODE, &mode) < 0) 
	{
		
		return err;
	}
	
	/* If we have a palletized mode, set the CLUT to 8-bit 3:3:2.
	** FIXME: There are additional palletized modes that need handling.
	*/
	if (depth == 8)
	{
		
		if (err = ioctl(fd, RAMDAC_SETCLUT, &palette_332) < 0) 
		{
			
			printf ("Error setting CLUT\n");
			exit(0);
		}
	}

	return 0;
}

/* Draw a pixel into the framebuffer.
** FIXME: Needs to handle 1- and 4-bit depths also.
*/
void putpixel (char *framebuffer, int xres, int yres, int x, int y, 
	       int depth, unsigned int color)
{
	uint8 *ptr8;
	uint16 *ptr16;
	uint32 *ptr32;

	/* Set the pixel as appropriate for the screen resolution and depth 
	*/
	if (depth == 8)
	{
		
		ptr8 = framebuffer + (xres * y) + x;
		*(framebuffer + (xres * y) + x) = color;
	}
	
	if (depth == 15 || depth == 16)
	{
		
		ptr16 = framebuffer + (xres * y * 2) + (x * 2);
		*ptr16 = (color & 0xFFFF);
	}

	if (depth == 24)
	{

		ptr8 = framebuffer + (xres * y * 3) + (x * 3);
		ptr8[0] = color & 0xFF;
		ptr8[1] = (color >> 8) & 0xFF;
		ptr8[2] = (color >> 16) & 0xFF;
	}
	
	if (depth == 32)
	{
		
		ptr32 = framebuffer + (xres * y * 4) + (x * 4);
		*ptr32 = color;
	}
}

/* Draw four colored bars on the screen as apporopriate for the given
** color depth.
*/
void draw_colorbars(char *framebuffer, int xres, int yres, int depth)
{
	float val;
	int i, j, r, g, b, w, c;
	
	for (i = 0; i < yres / 4; i++) /* Each bar is 1/4 screen height */
	{
		
		val = 0.0;
		for (j = 0; j < xres; j++) /* Step across the screen */
		{
			
			/* FIXME: Need to handle 1 and 4-bit depths */
			
			c = val;
			
			/* Just traverse the CLUT for palletized modes */
			if (depth == 8)
			{
				
				r = g = b = w = val = val + 1;
			}

			if (depth == 15) 
			{
				
				w = (c & 0x1f) + ((c & 0x1f) << 5) + 
				((c & 0x1f) << 10);
				b = c & 0x1f;
				g = (c & 0x1f) << 5;
				r = (c & 0x1f) << 10;
				val = val + 32.0 / xres;
			}
			
			if (depth == 16) 
			{
				
				w = ((c >> 1) & 0x1f) + ((c & 0x3f) << 5) + 
				(((c >> 1) & 0x1f) << 11);
				b = (c >> 1) & 0x1f;
				g = (c & 0x3f) << 5;
				r = ((c >> 1) & 0x1f) << 11;
				val = val + 64.0 / xres;
			}
			
			if (depth == 24 || depth == 32)
			{
				w = (c & 0xff) + ((c & 0xff) << 8) + 
				((c & 0xff) << 16);
				b = c & 0xff;
				g = (c & 0xff) << 8;
				r = (c & 0xff) << 16;
				val = val + 256.0 / xres;
			}
			
			putpixel(framebuffer, xres, yres, j, i, 
				 depth, w);
			putpixel(framebuffer, xres, yres, j, 
				 i + (yres / 4), depth, r);
			putpixel(framebuffer, xres, yres, j, i + 
				 ((yres / 4) * 2), depth, g);
			putpixel(framebuffer, xres, yres, j, 
				 i + ((yres / 4) * 3), depth, b);
		}
	}
}

/* Block until a KGI event is recieved.
*/
void wait_for_event(int fd)
{
	int size, err;
	fd_set fds;
	ggi_event *ev;
	char buffer[1024], *bp;
	
	FD_ZERO(&fds);		/* Zero out file descriptor set */
	FD_SET(fd, &fds);	/* Add our file descriptor to the set */
	
	/* Loop until a key is pressed.  If multiple events are queued,   */
	/* all of them will be read in and evaluated.                     */ 
	do
	{
		/* Block until /dev/graphic has something to tell us */
		err = select(fd + 1, &fds, NULL, NULL, NULL);

		/* Read in all queued events.  Currently, it is unlikely */
		/* that more than one event will ever be waiting, but we */
		/* read up to 1K from the device anyway and only look at */
		/* The first event.                                      */
		size = read(fd, buffer, 1024);
		
		ev = (ggi_event *)buffer;

	} while (ev->any.type != evKeyPress);
}

/* Returns a pointer to the linear framebuffer.
*/
char *mmap_framebuffer(int fd, int xres, int yres, int depth)
{
	int mmap_flags = 0, fbsize;
	char *framebuffer;
	
	fbsize = xres * yres;
	
	switch (depth) 
	{
	case 1: 
		fbsize /= 8; 
		break;
	case 4: 
		fbsize /= 2; 
		break;
	case 8: 
		break;
	case 15: 
		fbsize *= 2; 
		break;
	case 16: 
		fbsize *= 2; 
		break;
	case 24: 
		fbsize *= 3; 
		break; 
	case 32: 
		fbsize *= 4; 
		break;
	}

	/* Round fbsize up to the nearest 64K.  Necessary? */
	fbsize = ((fbsize >> 16) + ((fbsize & 0xffff) ? 1 : 0)) << 16; 
	
	/* Round fbsize up to nearest power of two divisor of 2MB. */
	/* The current KGI mmap() scheme will reject any mappings */
	/* That do not meet this criterion. */
	if (fbsize > 0x10000 && fbsize < 0x20000) 
	{
		
		fbsize = 0x20000;
	}
	else if (fbsize > 0x20000 && fbsize < 0x40000) 
	{
		
		fbsize = 0x40000;
	}
	else if (fbsize > 0x40000 && fbsize < 0x80000) 
	{
		
		fbsize = 0x80000;
	}
	else if (fbsize > 0x80000 && fbsize < 0x100000) 
	{
		
		fbsize = 0x100000;
	}
	else if (fbsize > 0x100000 && fbsize < 0x200000) 
	{
		
		fbsize = 0x200000;
	} 

	/* Everyone gets these */
	mmap_flags = MMAP_TYPE_MMIO | MMAP_PER_REGION_TYPE;
	
	/* Handle VGA ModeX-style planar mapping */
	/* FIXME: Non-ModeX modes can have these resolutions too! */
	if ((xres == 360 && depth == 8)
	{
		
		mmap_flags |= MMAP_IBM_VGA_MODEX;
		fbsize = 0x100000; /* This must be wrong, VGA only has 256K */
	}
	else
	{
		
		if (depth == 4) /* Another wierd mapping scheme */
		{
			
			mmap_flags |= MMAP_IBM_VGA_4;
			fbsize = 0x100000; /* This is wrong too */
		}
		else /* Standard linear framebuffer */
		{
			
			mmap_flags |= MMAP_FRAMEBUFFER;
		}
	}

	/* Try to get the memory-mapped linear framebuffer */
	if ((framebuffer = mmap((caddr_t)0, fbsize, PROT_READ | PROT_WRITE,
				MAP_SHARED, fd, mmap_flags)) == (void *)-1)
	{

		return (void *)-1;
	}

	/* Clear the framebuffer */
	memset(framebuffer, 0, fbsize); 
	
	return framebuffer;
}


/* A lot of stuff in here needs to be broken out into separate functions */
void main(int argc, char *argv[]) {
	int i, xres, yres, depth, type, err;
	int ggi_fd;
	int fbsize;
	char *framebuffer;
  
	if (argc != 4) 
	{
		
		printf("Usage: testvid <width> <height> <depth>\n");
		exit(0);
	}
		  
	xres = atoi(argv[1]);
	yres = atoi(argv[2]);
	depth = atoi(argv[3]);
  
	/* Open the KGI device so we can do stuff */
	if ((ggi_fd = open("/dev/graphic", O_RDWR)) == -1)
	{
		
		perror("Error opening /dev/graphic!");
		exit(0);
	}

	/* Try to set the mode given on the command line */
	if (setmode(ggi_fd, xres, yres, xres, yres, depth) != 0)
	{
		
		printf("Error setting mode!\n");
		exit(0);
	}
	
	if ((framebuffer = mmap_framebuffer(ggi_fd, xres, yres, depth)) == (void *)-1)
	{
		
		printf("Error mapping the framebuffer!\n");
		exit(0);
	}
		
	/* Draw the colorbars */
	draw_colorbars(framebuffer, xres, yres, depth);
	
	/* Wait for an event */
	wait_for_event(ggi_fd);
	
	/* We're done, so close everything down and exit. */

	close(ggi_fd);
	exit(0);
}
